@@ -36,7 +36,8 @@
 #define FLAG_NOBOOL	0x04
 #define FLAG_BOOLEAN	0x08
 #define FLAG_NOSTDIN	0x10
-#define FLAG_MASK	(FLAG_ARRAY | FLAG_PRETTY | FLAG_NOBOOL | FLAG_BOOLEAN | FLAG_NOSTDIN)
+#define FLAG_SKIPNULLS	0x20
+#define FLAG_MASK	(FLAG_ARRAY | FLAG_PRETTY | FLAG_NOBOOL | FLAG_BOOLEAN | FLAG_NOSTDIN | FLAG_SKIPNULLS)
 
 /* Size of buffer blocks for pipe slurping */
 #define SLURP_BLOCK_SIZE 4096
@@ -100,12 +101,6 @@ void json_copy_to_object(JsonNode * obj, JsonNode * object_or_array, int clobber
 	}
 }
 
-const char *maybe_stdin(const char* filename)
-{
-	return strcmp(filename, "-") ? filename : "/dev/fd/0";
-}
-		
-
 char *slurp_file(const char* filename, size_t *out_len, bool fold_newlines)
 {
 	char *buf;
@@ -113,8 +108,10 @@ char *slurp_file(const char* filename, size_t *out_len, bool fold_newlines)
 	int ch;
 	off_t buffer_len;
 	FILE *fp;
+	bool use_stdin = strcmp(filename, "-") == 0;
 
-	if ((fp = fopen(maybe_stdin(filename), "r")) == NULL) {
+	if (use_stdin) fp = stdin;
+	else if ((fp = fopen(filename, "r")) == NULL) {
 		perror(filename);
 		errx(1, "Cannot open %s for reading", filename);
 	}
@@ -142,7 +139,7 @@ char *slurp_file(const char* filename, size_t *out_len, bool fold_newlines)
 			buf[i++] = ch;
 		}
 	}
-        fclose(fp);
+	if (!use_stdin) fclose(fp);
 	buf[i] = 0;
 	*out_len = i;
 	return (buf);
@@ -224,7 +221,7 @@ JsonNode *vnode(char *str, int flags)
 	JsonTag type = flags_to_tag(flags);
 
 	if (strlen(str) == 0) {
-		return jo_mknull(type);
+		return (flags & FLAG_SKIPNULLS) ? (JsonNode *)NULL : jo_mknull(type);
 	}
 
 	/* If str begins with a double quote, keep it a string */
@@ -339,13 +336,15 @@ JsonNode *boolnode(char *str)
 
 int usage(char *prog)
 {
-	fprintf(stderr, "Usage: %s [-a] [-B] [-d keydelim] [-p] [-e] [-v] [-V] [--] [-s|-n|-b] [word...]\n", prog);
+	fprintf(stderr, "Usage: %s [-a] [-B] [-D] [-d keydelim] [-p] [-e] [-n] [-v] [-V] [-f file] [--] [-s|-n|-b] [word...]\n", prog);
 	fprintf(stderr, "\tword is key=value or key@value\n");
 	fprintf(stderr, "\t-a creates an array of words\n");
-	fprintf(stderr, "\t-B disable boolean true/false\n");
+	fprintf(stderr, "\t-B disable boolean true/false/null detection\n");
+	fprintf(stderr, "\t-D deduplicate object keys\n");
 	fprintf(stderr, "\t-d key will be object path separated by keydelim\n");
+	fprintf(stderr, "\t-f load file as JSON object or array\n");
 	fprintf(stderr, "\t-p pretty-prints JSON on output\n");
-	fprintf(stderr, "\t-e if stdin is empty do not wait for input quit\n");
+	fprintf(stderr, "\t-e quit if stdin is empty do not wait for input\n");
 	fprintf(stderr, "\t-s coerce type guessing to string\n");
 	fprintf(stderr, "\t-b coerce type guessing to bool\n");
 	fprintf(stderr, "\t-n coerce type guessing to number\n");
@@ -365,12 +364,14 @@ int usage(char *prog)
  *   *baseop -> object node in which caller should insert "value"
  */
 
-bool resolve_nested(int flags, char **keyp, char key_delim, char *value, JsonNode **baseop)
+bool resolve_nested(int flags, char **keyp, char key_delim, JsonNode *value, JsonNode **baseop)
 {
 	char *member = NULL, *bo, *bc, *so;		/* bracket open, close, sub-object */
 	JsonNode *op;
 	int found = false;
 
+	(void)flags;
+
 	if (key_delim) {
 		/* First construct nested object */
 		while ((so = strchr(*keyp, key_delim)) != NULL) {
@@ -411,17 +412,9 @@ bool resolve_nested(int flags, char **keyp, char key_delim, char *value, JsonNod
 		}
 
 		if (member == NULL) {		/* we're doing an array */
-			if (flags & FLAG_BOOLEAN) {
-				json_append_element(op, boolnode(value));
-			} else {
-				json_append_element(op, vnode(value, flags));
-			}
+			json_append_element(op, value);
 		} else {			/* we're doing an object */
-			if (flags & FLAG_BOOLEAN) {
-				json_append_member(op, member, boolnode(value));
-			} else {
-				json_append_member(op, member, vnode(value, flags));
-			}
+			json_append_member(op, member, value);
 		}
 
 		if (!found) {
@@ -457,7 +450,8 @@ int member_to_object(JsonNode *object, int flags, char key_delim, char *kv)
 		}
 
 		*r = 0;		/* Chop at ":=" */
-		json_append_member(object, kv, o);
+		if (!resolve_nested(flags, &kv, key_delim, o, &object))
+			json_append_member(object, kv, o);
 		return (0);
 	}
 
@@ -465,20 +459,20 @@ int member_to_object(JsonNode *object, int flags, char key_delim, char *kv)
 		return (-1);
 	}
 
-
+	JsonNode *val;
 	if (p) {
-		if (p) {
-			*p = 0;
+		*p = 0;
+		val = vnode(p+1, flags);
 
-			if (!resolve_nested(flags, &kv, key_delim, p+1, &object))
-				json_append_member(object, kv, vnode(p+1, flags));
-		}
+		if (!resolve_nested(flags, &kv, key_delim, val, &object))
+			json_append_member(object, kv, val);
 	} else {
 		if (q) {
 			*q = 0;
+			val = boolnode(q+1);
 
-			if (!resolve_nested(flags | FLAG_BOOLEAN, &kv, key_delim, q+1, &object))
-				json_append_member(object, kv, boolnode(q+1));
+			if (!resolve_nested(flags | FLAG_BOOLEAN, &kv, key_delim, val, &object))
+				json_append_member(object, kv, val);
 		}
 	}
 	return (0);
@@ -567,9 +561,9 @@ char* locale_from_utf8(const char *utf8, size_t len)
 # define locale_free(p) free(p)
 #else
 # define utf8_from_locale(p, l) (p)
-# define utf8_free(p)
+# define utf8_free(p) do {} while (0)
 # define locale_from_utf8(p, l) (p)
-# define locale_free(p)
+# define locale_free(p) do {} while (0)
 #endif
 
 char *stringify(JsonNode *json, int flags)
@@ -602,6 +596,8 @@ int main(int argc, char **argv)
 	int c, key_delim = 0;
 	bool showversion = false;
 	char *kv, *js_string, *progname, buf[BUFSIZ], *p;
+	char *in_file = NULL, *in_str;
+	size_t in_len = 0;
 	int ttyin = isatty(fileno(stdin)), ttyout = isatty(fileno(stdout));
 	int flags = 0;
 	JsonNode *json, *op;
@@ -614,7 +610,7 @@ int main(int argc, char **argv)
 
 	progname = (progname = strrchr(*argv, '/')) ? progname + 1 : *argv;
 
-	while ((c = getopt(argc, argv, "aBd:hpevV")) != EOF) {
+	while ((c = getopt(argc, argv, "aBDd:f:hpenvV")) != EOF) {
 		switch (c) {
 			case 'a':
 				flags |= FLAG_ARRAY;
@@ -622,9 +618,15 @@ int main(int argc, char **argv)
 			case 'B':
 				flags |= FLAG_NOBOOL;
 				break;
+			case 'D':
+				json_dedup_members(true);
+				break;
 			case 'd':
 				key_delim = optarg[0];
 				break;
+			case 'f':
+				in_file = optarg;
+				break;
 			case 'h':
 				usage(progname);
 				return (0);
@@ -634,6 +636,9 @@ int main(int argc, char **argv)
 			case 'e':
 				flags |= FLAG_NOSTDIN;
 				break;
+			case 'n':
+				flags |= FLAG_SKIPNULLS;
+				break;
 			case 'v':
 				printf("jo %s\n", PACKAGE_VERSION);
 				exit(0);
@@ -653,7 +658,26 @@ int main(int argc, char **argv)
 	argv += optind;
 
 	pile = json_mkobject();
-	json = (flags & FLAG_ARRAY) ? json_mkarray() : json_mkobject();
+	if (in_file != NULL) {
+		if ((in_str = slurp_file(in_file, &in_len, false)) == NULL) {
+			errx(1, "Error reading file %s", in_file);
+		}
+		json = json_decode(in_str);
+		if (json) {
+			switch (json->tag) {
+				case JSON_ARRAY:
+					flags |= FLAG_ARRAY;
+					break;
+				case JSON_OBJECT:
+					break;
+				default:
+					errx(1, "Input JSON not an array or object: %s", stringify(json, flags));
+			}
+		} else
+			json = (flags & FLAG_ARRAY) ? json_mkarray() : json_mkobject();
+	} else {
+		json = (flags & FLAG_ARRAY) ? json_mkarray() : json_mkobject();
+	}
 
 	if (argc == 0) {
 		if (flags & FLAG_NOSTDIN) {
