@@ -34,6 +34,18 @@
 		exit(EXIT_FAILURE);                     \
 	} while (0)
 
+#ifdef _WIN32
+# define failx(e, n, f, ...)	if (!(e)) {	\
+		fprintf(stderr, "JSON_ERR: " f, __VA_ARGS__);	\
+		exit(n);	\
+	}
+#else
+# include <err.h>
+# define failx(e, n, f, ...)	if (!(e)) {	\
+		errx(n, "JSON_ERR: " f, __VA_ARGS__);	\
+	}
+#endif
+
 /* Sadly, strdup is not portable. */
 static char *json_strdup(const char *str)
 {
@@ -271,7 +283,12 @@ static int utf8_write_char(js_uchar_t unicode, char *out)
 {
 	unsigned char *o = (unsigned char*) out;
 	
-	assert(unicode <= 0x10FFFF && !(unicode >= 0xD800 && unicode <= 0xDFFF));
+	failx(
+		unicode <= 0x10FFFF && !(unicode >= 0xD800 && unicode <= 0xDFFF),
+		1,
+		"Illegal Unicode codepoint 0x%08X found",
+		unicode
+	);
 
 	if (unicode <= 0x7F) {
 		/* U+0000..U+007F */
@@ -323,7 +340,12 @@ static void to_surrogate_pair(js_uchar_t unicode, uint16_t *uc, uint16_t *lc)
 {
 	js_uchar_t n;
 	
-	assert(unicode >= 0x10000 && unicode <= 0x10FFFF);
+	failx(
+		unicode >= 0x10000 && unicode <= 0x10FFFF,
+		1,
+		"Cannot construct UTF-16 surrogate pair for Unicode codepoint 0x%08X",
+		unicode
+	);
 	
 	n = unicode - 0x10000;
 	*uc = ((n >> 10) & 0x3FF) | 0xD800;
@@ -357,8 +379,11 @@ static int write_hex16(char *out, uint16_t val);
 static JsonNode *mknode(JsonTag tag);
 static void append_node(JsonNode *parent, JsonNode *child);
 static void prepend_node(JsonNode *parent, JsonNode *child);
+static void insert_node(JsonNode *parent, JsonNode *child);
 static void append_member(JsonNode *object, char *key, JsonNode *value);
 
+static void (*append_member_node_fn)(JsonNode *parent, JsonNode *child) = append_node;
+
 /* Assertion-friendly validity checks */
 static bool tag_is_valid(unsigned int tag);
 static bool number_is_valid(const char *num);
@@ -540,6 +565,7 @@ JsonNode *json_mkobject(void)
 
 static void append_node(JsonNode *parent, JsonNode *child)
 {
+	if (!child) return;
 	child->parent = parent;
 	child->prev = parent->children.tail;
 	child->next = NULL;
@@ -553,6 +579,7 @@ static void append_node(JsonNode *parent, JsonNode *child)
 
 static void prepend_node(JsonNode *parent, JsonNode *child)
 {
+	if (!child) return;
 	child->parent = parent;
 	child->prev = NULL;
 	child->next = parent->children.head;
@@ -564,15 +591,52 @@ static void prepend_node(JsonNode *parent, JsonNode *child)
 	parent->children.head = child;
 }
 
+static void insert_node(JsonNode *parent, JsonNode *child)
+{
+	if (!child) return;
+	JsonNode *this = parent->children.head;
+
+	while (this != NULL && strcmp(this->key, child->key))
+		this = this->next;
+
+	if (this != NULL)
+	{
+		/* we found a matching key, insert before this node */
+		if (this->prev == NULL)
+		{
+			prepend_node(parent, child);
+		}
+		else
+		{
+			child->parent = parent;
+			child->next = this->next;
+			child->prev = this->prev;
+			this->prev->next = child;
+			this->prev = child;
+		}
+		json_delete(this);
+	}
+	else
+		append_node(parent, child);
+}
+
 static void append_member(JsonNode *object, char *key, JsonNode *value)
 {
+	if (!value) return;
 	value->key = key;
-	append_node(object, value);
+	append_member_node_fn(object, value);
 }
 
 void json_append_element(JsonNode *array, JsonNode *element)
 {
-	assert(array->tag == JSON_ARRAY);
+	if (!element) return;
+	failx(
+		array->tag == JSON_ARRAY,
+		1,
+		"Cannot append %s to non-array %s",
+		json_encode(element),
+		json_encode(array)
+	);
 	assert(element->parent == NULL);
 	
 	append_node(array, element);
@@ -580,7 +644,14 @@ void json_append_element(JsonNode *array, JsonNode *element)
 
 void json_prepend_element(JsonNode *array, JsonNode *element)
 {
-	assert(array->tag == JSON_ARRAY);
+	if (!element) return;
+	failx(
+		array->tag == JSON_ARRAY,
+		1,
+		"Cannot append %s to non-array %s",
+		json_encode(element),
+		json_encode(array)
+	);
 	assert(element->parent == NULL);
 	
 	prepend_node(array, element);
@@ -588,7 +659,15 @@ void json_prepend_element(JsonNode *array, JsonNode *element)
 
 void json_append_member(JsonNode *object, const char *key, JsonNode *value)
 {
-	assert(object->tag == JSON_OBJECT);
+	if (!value) return;
+	failx(
+		object->tag == JSON_OBJECT,
+		1,
+		"Cannot add {\"%s\":%s} to non-object %s",
+		key,
+		json_encode(value),
+		json_encode(object)
+	);
 	assert(value->parent == NULL);
 	
 	append_member(object, json_strdup(key), value);
@@ -596,13 +675,26 @@ void json_append_member(JsonNode *object, const char *key, JsonNode *value)
 
 void json_prepend_member(JsonNode *object, const char *key, JsonNode *value)
 {
-	assert(object->tag == JSON_OBJECT);
+	if (!value) return;
+	failx(
+		object->tag == JSON_OBJECT,
+		1,
+		"Cannot add {\"%s\":%s} to non-object %s",
+		key,
+		json_encode(value),
+		json_encode(object)
+	);
 	assert(value->parent == NULL);
 	
 	value->key = json_strdup(key);
 	prepend_node(object, value);
 }
 
+void json_dedup_members(bool b)
+{
+	append_member_node_fn = b ? insert_node : append_node;
+}
+
 void json_remove_from_parent(JsonNode *node)
 {
 	JsonNode *parent = node->parent;
@@ -1181,9 +1273,9 @@ void emit_string(SB *out, const char *str)
 #endif
 						b += 6;
 					} else {
-						*b++ = 0xEF;
-						*b++ = 0xBF;
-						*b++ = 0xBD;
+						*b++ = (char)0xEF;
+						*b++ = (char)0xBF;
+						*b++ = (char)0xBD;
 					}
 					s++;
 				} else if (c < 0x1F || (c >= 0x80 && escape_unicode)) {
